==> guildai-0.8.2-py2.py3-none-any.whl: guild/tests/dependencies-2.md <==
# Dependencies 2

These tests cover the various alternative dependency definition
schemes that were added to simplify dependency definitions.

The primary goal of this simplification is to support inline
dependencies that do not require separate resource definitions.

We'll use a helper to print depepdencies:

    >>> def print_deps(deps):
    ...     for dep in deps:
    ...         print(dep)
    ...         if dep.spec is not None:
    ...             print("  spec: %s" % dep.spec)
    ...         if dep.description is not None:
    ...             print("  description: %s" % (dep.description or "''"))
    ...         if dep.inline_resource:
    ...             print("  inline-resource: ", end="")
    ...             pprint(dep.inline_resource._data)

## Inline string

An inline string is the traditional way to express a dependency on a
resource.

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires: prepare
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef 'prepare'>
      spec: prepare
      description: ''

A list of strings is treated in the same manner:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     - prepare-1
    ...     - prepare-2
    ...     - prepare-3
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef 'prepare-1'>
      spec: prepare-1
      description: ''
    <guild.guildfile.OpDependencyDef 'prepare-2'>
      spec: prepare-2
      description: ''
    <guild.guildfile.OpDependencyDef 'prepare-3'>
      spec: prepare-3
      description: ''

## Inline dict - resource spec with description

An inline dict can be either a resource reference or an inline
resource. If the dict contains a `resource` attribute, it's treated as
a resource spec:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     resource: data
    ...     description: required data
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef 'data'>
      spec: data
      description: required data

## Inline dict - implicit source

If the user provides an inline dict that does not contain a `sources`
attribute, it is treated as an implicit resource source:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     url: http://my.co/stuff.gz
    ...     sha256: abc123
    ...     unpack: no
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef http://my.co/stuff.gz>
      inline-resource: {'sources': [{'sha256': 'abc123',
                  'unpack': False,
                  'url': 'http://my.co/stuff.gz'}]}

## Inline dict - full resource definition

If the inline dict contains a `sources` attribute, it is treated as a
complete resource definition:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     target-path: data
    ...     sources:
    ...       - url: http://my.co/stuff.gz
    ...         sha256: abc123
    ...         unpack: no
    ...       - url: http://my.co/more-suff.tar
    ...         sha256: dev456
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef http://my.co/stuff.gz,http://my.co/more-suff.tar>
      inline-resource: {'sources': [{'sha256': 'abc123',
                  'unpack': False,
                  'url': 'http://my.co/stuff.gz'},
                 {'sha256': 'dev456', 'url': 'http://my.co/more-suff.tar'}],
     'target-path': 'data'}

## Named inline resources

A resource `name` may be provided for any full resource definition:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     name: data
    ...     sources:
    ...       - url: http://my.co/stuff.gz
    ...       - url: http://my.co/more-suff.tar
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef data>
      inline-resource: {'name': 'data',
     'sources': [{'url': 'http://my.co/stuff.gz'},
                 {'url': 'http://my.co/more-suff.tar'}]}

## Named and unnamed operation resources

Inline requirements that define an operation but not a name are
implicitly named by their operation.

    >>> gf = guildfile.for_string("""
    ... op1:
    ...   requires:
    ...     - operation: foo
    ... op2:
    ...   requires:
    ...     - operation: foo
    ...       name: FOO
    ... """)

    >>> gf.default_model["op1"].dependencies[0].inline_resource.name
    'foo'

    >>> gf.default_model["op2"].dependencies[0].inline_resource.name
    'FOO'

## Inline list

An inline list can be used to list source. Each list item is converted
to a dependency where `sources` contains a list containing the item.

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     - url: http://my.co/stuff.gz
    ...     - operation: foo
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef http://my.co/stuff.gz>
      inline-resource: {'sources': [{'url': 'http://my.co/stuff.gz'}]}
    <guild.guildfile.OpDependencyDef foo>
      inline-resource: {'sources': [{'operation': 'foo'}]}

## Source paths

As of 0.6.2, a source may contain a path, which is in addition to any
path defined in the source resource. A source path is appended to a
resource path, if a resource path is defined.

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     - operation: foo
    ...       select: .+\.txt
    ...       target-path: bar
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies) # doctest: -NORMALIZE_PATHS
    <guild.guildfile.OpDependencyDef foo>
      inline-resource: {'sources': [{'operation': 'foo', 'select': '.+\\.txt', 'target-path': 'bar'}]}

    <guild.guildfile.OpDependencyDef foo>
      inline-resource: {'sources': [{'operation': 'foo',
                                     'path': 'bar',
                                     'select': '.+\\.txt'}]}

## Project tests

The sample project `inline-resources` contains a variety of inline
resource definitions. We can test the resolution of all of the
resources by running the `test-all` operation, which serves as a
validation of the expected results.

Note we skip running the operations on Windows because they use POSIX
executables.

    >>> project = Project(sample("projects", "inline-resources"))

### Force run on dependency error

When a dependency cannot be met, Guild stops the run with an error
message.

The `print-msg-2` operation requires `print-msg`, which we don't have.

    >>> project.list_runs()
    []

`print-msg-2` fails with an error message.

    >>> project.run("print-msg-2")
    WARNING: cannot find a suitable run for required resource 'print-msg'
    Resolving print-msg
    guild: run failed because a dependency was not met: could not resolve
    'operation:print-msg' in print-msg resource: no suitable run for print-msg
    <exit 1>

We can force a run even in such cases using `force_deps`.

    >>> project.run("print-msg-2", force_deps=True)
    WARNING: cannot find a suitable run for required resource 'print-msg'
    Resolving print-msg
    WARNING: a dependency was not met: could not resolve 'operation:print-msg'
    in print-msg resource: no suitable run for print-msg
    cat: no such file 'msg.txt'
    <exit 1>

The same behavior applies to staging runs. We use `missing-file-dep`
to test this, which relies on a file that doesn't exist.

    >>> project.run("missing-file-dep", stage=True)
    Resolving file:missing.txt
    guild: run failed because a dependency was not met: could not resolve
    'file:missing.txt' in file:missing.txt resource: cannot find source file 'missing.txt'
    <exit 1>

Using `force_deps` we can bypass the check and stage the run anyway.

    >>> project.run("missing-file-dep", stage=True, force_deps=True)
    Resolving file:missing.txt
    WARNING: a dependency was not met: could not resolve 'file:missing.txt' in
    file:missing.txt resource: cannot find source file 'missing.txt'
    Resolving file:msg.txt
    missing-file-dep staged as ...
    To start the operation, use 'guild run --start ...'

Dependencies that can be resolved are resolved. `missing-file-dep`
also requires `msg.txt`, which does exist and can be resooved.

    >>> dir(project.list_runs()[0].dir)
    ['.guild', 'msg.txt']

### test-all

    >>> project.run("test-all")
    INFO: [guild] running print-msg: print-msg
    Resolving file:msg.txt
    hola
    INFO: [guild] checking run ... files 'msg.txt'
    INFO: [guild] comparing run ... file msg.txt to .../inline-resources/msg.txt
    INFO: [guild] 1 of 1 checks passed
    INFO: [guild] running print-msg-2: print-msg-2
    Resolving print-msg
    Using run ... for print-msg resource
    hola
    INFO: [guild] checking run ... files 'msg.txt'
    INFO: [guild] comparing run ... file msg.txt to .../inline-resources/msg.txt
    INFO: [guild] 1 of 1 checks passed
    INFO: [guild] running print-msg-3: print-msg-3
    Resolving file:msg.txt
    Resolving file:msg-2.txt
    hola
    adios
    INFO: [guild] checking run ... files 'msg-1.txt'
    INFO: [guild] comparing run ... file msg-1.txt to .../inline-resources/msg.txt
    INFO: [guild] checking run ... files 'msg-2.txt'
    INFO: [guild] comparing run ... file msg-2.txt to .../inline-resources/msg-2.txt
    INFO: [guild] 2 of 2 checks passed
    INFO: [guild] running print-msg-4: print-msg-4
    Resolving file:msg.txt,file:msg-2.txt
    hola
    adios
    INFO: [guild] checking run ... files 'messages/msg.txt'
    INFO: [guild] comparing run ... file messages/msg.txt to .../inline-resources/msg.txt
    INFO: [guild] checking run ... files 'messages/msg-2.txt'
    INFO: [guild] comparing run ... file messages/msg-2.txt to .../inline-resources/msg-2.txt
    INFO: [guild] 2 of 2 checks passed

==> guildai-0.9.0rc1-py2.py3-none-any.whl: guild/tests/dependencies-2.md <==
# Dependencies 2

These tests cover the various alternative dependency definition
schemes that were added to simplify dependency definitions.

The primary goal of this simplification is to support inline
dependencies that do not require separate resource definitions.

We use a helper to print depepdencies:

    >>> def print_deps(deps):
    ...     for dep in deps:
    ...         print(dep)
    ...         if dep.spec is not None:
    ...             print("  spec: %s" % dep.spec)
    ...         if dep.description is not None:
    ...             print("  description: %s" % (dep.description or "''"))
    ...         if dep.inline_resource:
    ...             print("  inline-resource: ", end="")
    ...             pprint(dep.inline_resource._data)

## Inline string

An inline string is the traditional way to express a dependency on a
resource.

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires: prepare
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef 'prepare'>
      spec: prepare
      description: ''

A list of strings is treated in the same manner:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     - prepare-1
    ...     - prepare-2
    ...     - prepare-3
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef 'prepare-1'>
      spec: prepare-1
      description: ''
    <guild.guildfile.OpDependencyDef 'prepare-2'>
      spec: prepare-2
      description: ''
    <guild.guildfile.OpDependencyDef 'prepare-3'>
      spec: prepare-3
      description: ''

## Inline dict - resource spec with description

An inline dict can be either a resource reference or an inline
resource. If the dict contains a `resource` attribute, it's treated as
a resource spec:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     resource: data
    ...     description: required data
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef 'data'>
      spec: data
      description: required data

## Inline dict - implicit source

If the user provides an inline dict that does not contain a `sources`
attribute, it is treated as an implicit resource source:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     url: http://my.co/stuff.gz
    ...     sha256: abc123
    ...     unpack: no
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef http://my.co/stuff.gz>
      inline-resource: {'sources': [{'sha256': 'abc123',
                  'unpack': False,
                  'url': 'http://my.co/stuff.gz'}]}

## Inline dict - full resource definition

If the inline dict contains a `sources` attribute, it is treated as a
complete resource definition:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     target-path: data
    ...     sources:
    ...       - url: http://my.co/stuff.gz
    ...         sha256: abc123
    ...         unpack: no
    ...       - url: http://my.co/more-suff.tar
    ...         sha256: dev456
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef http://my.co/stuff.gz,http://my.co/more-suff.tar>
      inline-resource: {'sources': [{'sha256': 'abc123',
                  'unpack': False,
                  'url': 'http://my.co/stuff.gz'},
                 {'sha256': 'dev456', 'url': 'http://my.co/more-suff.tar'}],
     'target-path': 'data'}

## Named inline resources

A resource `name` may be provided for any full resource definition:

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     name: data
    ...     sources:
    ...       - url: http://my.co/stuff.gz
    ...       - url: http://my.co/more-suff.tar
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef data>
      inline-resource: {'name': 'data',
     'sources': [{'url': 'http://my.co/stuff.gz'},
                 {'url': 'http://my.co/more-suff.tar'}]}

## Named and unnamed operation resources

Inline requirements that define an operation but not a name are
implicitly named by their operation.

    >>> gf = guildfile.for_string("""
    ... op1:
    ...   requires:
    ...     - operation: foo
    ... op2:
    ...   requires:
    ...     - operation: foo
    ...       name: FOO
    ... """)

    >>> gf.default_model["op1"].dependencies[0].inline_resource.resolving_name
    'operation:foo'

    >>> gf.default_model["op2"].dependencies[0].inline_resource.resolving_name
    'FOO'

## Inline list

An inline list can be used to list source. Each list item is converted
to a dependency where `sources` contains a list containing the item.

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     - url: http://my.co/stuff.gz
    ...     - operation: foo
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies)
    <guild.guildfile.OpDependencyDef http://my.co/stuff.gz>
      inline-resource: {'sources': [{'url': 'http://my.co/stuff.gz'}]}
    <guild.guildfile.OpDependencyDef operation:foo>
      inline-resource: {'sources': [{'operation': 'foo'}]}

## Source paths

As of 0.6.2, a source may contain a path, which is in addition to any
path defined in the source resource. A source path is appended to a
resource path, if a resource path is defined.

    >>> gf = guildfile.for_string("""
    ... train:
    ...   requires:
    ...     - operation: foo
    ...       select: .+\.txt
    ...       target-path: bar
    ... """)

    >>> print_deps(gf.default_model["train"].dependencies) # doctest: -NORMALIZE_PATHS
    <guild.guildfile.OpDependencyDef operation:foo>
      inline-resource: {'sources': [{'operation': 'foo',
                                     'select': '.+\\.txt',
                                     'target-path': 'bar'}]}

## Project tests

The sample project `inline-resources` contains a variety of inline resource
definitions. We can test the resolution of all of the resources by running the
`test-all` operation, which serves as a validation of the expected results.

    >>> use_project("inline-resources")

Some of operations require various `msg*.txt` files, which may be
available or not. The project is configured with `.guildignore` to
skip these files as source code to prevent incidentally resolving the
file by way of a source code copy.

    >>> cat(".guildignore")
    msg*.txt

### Force run on dependency error

When a dependency can't be met, Guild stops the run with an error
message.

The `print-msg-2` operation requires `print-msg`, which we don't have.

    >>> run("guild runs")
    <exit 0>

`print-msg-2` fails with an error message.

    >>> run("guild run print-msg-2 -y")
    WARNING: cannot find a suitable run for required resource 'operation:print-msg'
    Resolving operation:print-msg
    guild: run failed because a dependency was not met: could not resolve
    'operation:print-msg' in operation:print-msg resource: no suitable run for print-msg
    <exit 1>

We can force a run in this case using `--force-deps`.

    >>> run("guild run print-msg-2 --force-deps -y")
    WARNING: cannot find a suitable run for required resource 'operation:print-msg'
    Resolving operation:print-msg
    WARNING: a dependency was not met: could not resolve 'operation:print-msg'
    in operation:print-msg resource: no suitable run for print-msg
    cat: no such file 'msg.txt'
    <exit 1>

The same behavior applies to staging runs. We use `missing-file-dep`
to show this, which relies on a file that doesn't exist.

    >>> run("guild run missing-file-dep --stage -y")
    Resolving file:missing.txt
    guild: run failed because a dependency was not met: could not resolve
    'file:missing.txt' in file:missing.txt resource: cannot find source file 'missing.txt'
    <exit 1>

With `--force-deps` we can bypass the check and stage the run.

    >>> run("guild run missing-file-dep --stage --force-deps -y")
    Resolving file:missing.txt
    WARNING: a dependency was not met: could not resolve 'file:missing.txt' in
    file:missing.txt resource: cannot find source file 'missing.txt'
    Resolving file:msg.txt
    missing-file-dep staged as ...
    To start the operation, use 'guild run --start ...'

Dependencies that can be resolved are resolved. `missing-file-dep`
also requires `msg.txt`, which does exist and can be resolved.

    >>> run("guild ls -n")
    cat.py
    guild.yml
    msg.txt

### test-all

The operation `test-all` uses steps to run a series of
opertaions. Each step defines `check` criteria to verify the
resolution of files.

    >>> run("guild run test-all -y")  # doctest: +REPORT_UDIFF
    INFO: [guild] running print-msg: print-msg
    Resolving file:msg.txt
    hola
    INFO: [guild] checking run ... files 'msg.txt'
    INFO: [guild] comparing run ... file msg.txt to .../inline-resources/msg.txt
    INFO: [guild] 1 of 1 checks passed
    INFO: [guild] running print-msg-2: print-msg-2
    Resolving operation:print-msg
    Using run ... for operation:print-msg
    hola
    INFO: [guild] checking run ... files 'msg.txt'
    INFO: [guild] comparing run ... file msg.txt to .../inline-resources/msg.txt
    INFO: [guild] 1 of 1 checks passed
    INFO: [guild] running print-msg-3: print-msg-3
    Resolving file:msg.txt
    Resolving file:msg-2.txt
    hola
    adios
    INFO: [guild] checking run ... files 'msg-1.txt'
    INFO: [guild] comparing run ... file msg-1.txt to .../inline-resources/msg.txt
    INFO: [guild] checking run ... files 'msg-2.txt'
    INFO: [guild] comparing run ... file msg-2.txt to .../inline-resources/msg-2.txt
    INFO: [guild] 2 of 2 checks passed
    INFO: [guild] running print-msg-4: print-msg-4
    Resolving file:msg.txt,file:msg-2.txt
    hola
    adios
    INFO: [guild] checking run ... files 'messages/msg.txt'
    INFO: [guild] comparing run ... file messages/msg.txt to .../inline-resources/msg.txt
    INFO: [guild] checking run ... files 'messages/msg-2.txt'
    INFO: [guild] comparing run ... file messages/msg-2.txt to .../inline-resources/msg-2.txt
    INFO: [guild] 2 of 2 checks passed
